@@ -143,6 +143,24 @@ char	fliptab[8] = {
 	0x80,
 };
 
+int flip_error_corr(unsigned char *b, int crc)
+{
+  if (crc != 0) {
+    int i;
+    for (i = 0; i < SUBSIZE; i++) {
+      char	c;
+
+      c = fliptab[i%8];
+      u.b[i / 8] ^= c;
+      if ((crc = calcCRC(u.b, SUBSIZE/8)) == 0) {
+        return crc;
+      }
+      u.b[i / 8] ^= c;
+    }
+  }
+  return crc;
+}
+
 /* collect bytes from bit-channels */
 int getQ(unsigned char subs[])
 {
@@ -259,17 +277,8 @@ int getQ(unsigned char subs[])
 
   /* calculate CRC checksum */
   crc =  calcCRC(u.b, SUBSIZE/8);
-  if (crc != 0) {
-    for (i = 0; i < SUBSIZE; i++) {
-      char	c;
 
-      c = fliptab[i%8];
-      u.b[i / 8] ^= c;
-      if ((crc = calcCRC(u.b, SUBSIZE/8)) == 0)
-        break;
-      u.b[i / 8] ^= c;
-    }
-  }
+  crc = flip_error_corr(u.b, crc);
 
   return crc;
 }
@@ -284,12 +293,15 @@ unsigned char bin2bcd(unsigned p)
   return ((p/10)<<4)|(p%10);
 }
 
-#define DEBUG_HEUR 0
+static int crc_errors;
+
+#define DEBUG_HEUR 1
 int correct_heuristically(union u u2, unsigned char tr, unsigned char in,
-                          unsigned last_asec, unsigned last_rsec)
+                          unsigned last_asec, unsigned last_rsec, unsigned crcres)
 {
   int result = 0;
   int newtrack;
+  int transitionindex1;
 
   bcd2bin(&u2.q.track);
   bcd2bin(&u2.q.index);
@@ -300,10 +312,12 @@ int correct_heuristically(union u u2, unsigned char tr, unsigned char in,
   bcd2bin(&u2.q.asec);
   bcd2bin(&u2.q.aframe);
 
-if (DEBUG_HEUR) fprintf(stderr, "am %d:%2d.%02d,  rm %d:%2d.%02d, tr %d, in %d, last_asec %u, last_rsec %u, ltr %d, lin %d\n",
+if (DEBUG_HEUR) fprintf(stdout, "am %d:%2d.%02d, rm %d:%2d.%02d, tr %d, in %d, "
+"ctl_adr %x, zero %x, crc1 %02x, crc2 %02x, las %u, lrs %u, ltr %d, lin %d\n",
   u2.q.amin, u2.q.asec, u2.q.aframe,
   u2.q.min, u2.q.sec, u2.q.frame,
   u2.q.track, u2.q.index,
+  u2.q.ctrl_adr, u2.q.zero, u2.q.crc1, u2.q.crc2,
   last_asec, last_rsec,
   tr, in);
 
@@ -316,26 +330,27 @@ if (DEBUG_HEUR) fprintf(stderr, "am %d:%2d.%02d,  rm %d:%2d.%02d, tr %d, in %d,
     if (u2.q.aframe != (last_asec + 151) % 75) {
       u2.q.aframe = ((last_asec + 151) % 75);
       u.q.aframe = bin2bcd(u2.q.aframe);
-if (DEBUG_HEUR) fprintf(stderr, "corrected aframe to %x\n", u.q.aframe);
+if (DEBUG_HEUR) fprintf(stdout, "corrected aframe to %x\n", u.q.aframe);
       result = 1;
     }
     if (u2.q.asec != ((last_asec + 151) / 75) % 60) {
       u2.q.asec = (((last_asec + 151) / 75) % 60);
       u.q.asec = bin2bcd(u2.q.asec);
-if (DEBUG_HEUR) fprintf(stderr, "corrected asec to %x\n", u.q.asec);
+if (DEBUG_HEUR) fprintf(stdout, "corrected asec to %x\n", u.q.asec);
       result = 1;
     }
     if (u2.q.amin != ((last_asec + 151) / 75) / 60) {
       u2.q.amin = (((last_asec + 151) / 75) / 60);
       u.q.amin = bin2bcd(u2.q.amin);
-if (DEBUG_HEUR) fprintf(stderr, "corrected amin to %x\n", u.q.amin);
+if (DEBUG_HEUR) fprintf(stdout, "corrected amin to %x\n", u.q.amin);
       result = 1;
     }
   }
 
   newtrack = 0;
-  if (u2.q.frame == 0 && u2.q.min == 0 && u2.q.sec == 0 &&
-      u2.q.index == 1) {
+  if ((u2.q.track == tr + 1 && u2.q.index == 0) ||
+      (u2.q.track == tr + 1 && (u2.q.index == 1 || !(u2.q.frame || u2.q.sec || u2.q.min)))
+     ) {
     newtrack = 1;
   }
   /* check track */
@@ -344,66 +359,113 @@ if (DEBUG_HEUR) fprintf(stderr, "corrected amin to %x\n", u.q.amin);
       (newtrack && u2.q.track != tr + 1)) {
       u2.q.track = newtrack ? tr + 1 : tr;
       u.q.track = bin2bcd(u2.q.track);
-if (DEBUG_HEUR) fprintf(stderr, "corrected track to %x\n", u.q.track);
+if (DEBUG_HEUR) fprintf(stdout, "corrected track to %x\n", u.q.track);
       result = 1;
   }
+  if (u2.q.track == tr + 1 && u2.q.index == 0) {
+    newtrack = 1;
+  }
   /* check index */
   if (u2.q.index > 99U ||
       (!newtrack && (u2.q.index != in && u2.q.index != in + 1)) ||
-      (newtrack && u2.q.index != 1)) {
-    u2.q.index = newtrack ? 0 : in;
+      (newtrack && (u2.q.index != 0 && u2.q.index != 1))) {
+    u2.q.index = newtrack ? 0 : (in == 0 && last_rsec == 0) ? 1 : in;
     u.q.index = bin2bcd(u2.q.index);
-if (DEBUG_HEUR) fprintf(stderr, "corrected index to %x\n", u.q.index);
+if (DEBUG_HEUR) fprintf(stdout, "corrected index to %x\n", u.q.index);
       result = 1;
   }
+
+  if (u2.q.index == 1 && in == 0) {
+    transitionindex1 = 1;
+  } else transitionindex1 = 0;
+
   /* check relative position */
-  if (u2.q.frame > 74U ||
-      u2.q.min > 59U ||
-      u2.q.sec > 59U) {
-    /* relative position is dubious */
-    if (u2.q.frame != (last_rsec + 1) % 75) {
-      u2.q.frame = newtrack ? 0 : ((last_rsec + 1) % 75);
+
+  if (newtrack /*u2.q.index == 0 && in != 0*/) {
+    last_rsec = 60*75*u2.q.min + 75*u2.q.sec + u2.q.frame -1;
+  } else if (u2.q.index == 0 && in == 0) {
+    /* relative address counts backwards during index 0 */
+    if (u2.q.frame > 74U ||
+        u2.q.frame != (last_rsec - 1) % 75) {
+      u2.q.frame = ((last_rsec - 1) % 75);
       u.q.frame = bin2bcd(u2.q.frame);
-if (DEBUG_HEUR) fprintf(stderr, "corrected frame to %x\n", u.q.frame);
+if (DEBUG_HEUR) fprintf(stdout, "corrected frame to %x\n", u.q.frame);
       result = 1;
     }
-    if (u2.q.sec != ((last_rsec + 1) / 75) % 60) {
-      u2.q.sec = newtrack ? 0 : (((last_rsec + 1) / 75) % 60);
+    if (u2.q.sec > 59U ||
+        u2.q.sec != ((last_rsec - 1) / 75) % 60) {
+      u2.q.sec = (((last_rsec - 1) / 75) % 60);
       u.q.sec = bin2bcd(u2.q.sec);
-if (DEBUG_HEUR) fprintf(stderr, "corrected sec to %x\n", u.q.sec);
+if (DEBUG_HEUR) fprintf(stdout, "corrected sec to %x\n", u.q.sec);
       result = 1;
     }
-    if (u2.q.min != ((last_rsec + 1) / 75) / 60) {
-      u2.q.min = newtrack ? 0 : (((last_rsec + 1) / 75) / 60);
+    if (u2.q.min > 59U ||
+        u2.q.min != ((last_rsec - 1) / 75) / 60) {
+      u2.q.min = (((last_rsec - 1) / 75) / 60);
       u.q.min = bin2bcd(u2.q.min);
-if (DEBUG_HEUR) fprintf(stderr, "corrected min to %x\n", u.q.min);
+if (DEBUG_HEUR) fprintf(stdout, "corrected min to %x\n", u.q.min);
+      result = 1;
+    }
+  } else if (u2.q.index > 0) {
+    if (u2.q.frame > 74U ||
+        u2.q.frame != (last_rsec + 1) % 75) {
+      u2.q.frame = ((last_rsec + 1) % 75);
+      u.q.frame = bin2bcd(u2.q.frame);
+if (DEBUG_HEUR) fprintf(stdout, "2corrected frame to %x\n", u.q.frame);
+      result = 1;
+    }
+    if (u2.q.sec > 59U ||
+        u2.q.sec != ((last_rsec + 1) / 75) % 60) {
+      u2.q.sec = (((last_rsec + 1) / 75) % 60);
+      u.q.sec = bin2bcd(u2.q.sec);
+if (DEBUG_HEUR) fprintf(stdout, "2corrected sec to %x\n", u.q.sec);
+      result = 1;
+    }
+    if (u2.q.min > 59U ||
+        u2.q.min != ((last_rsec + 1) / 75) / 60) {
+      u2.q.min = (((last_rsec + 1) / 75) / 60);
+      u.q.min = bin2bcd(u2.q.min);
+if (DEBUG_HEUR) fprintf(stdout, "2corrected min to %x\n", u.q.min);
       result = 1;
     }
   }
+
+  if (u2.q.index == 1 && in == 0)
+    last_rsec = -1;
+
   if (((u2.q.ctrl_adr & 0x0f) != 1)) {
-if (DEBUG_HEUR) fprintf(stderr, "ctl_adr = %x,  zero = %x\n", u2.q.ctrl_adr, u2.q.zero);
+if (DEBUG_HEUR) fprintf(stdout, "ctl_adr = %x,  zero = %x\n", u2.q.ctrl_adr, u2.q.zero);
     u.q.ctrl_adr &= 0xf0;
     u.q.ctrl_adr |= 1;
-if (DEBUG_HEUR) fprintf(stderr, "corrected ctl_adr to %x\n", u.q.ctrl_adr);
+if (DEBUG_HEUR) fprintf(stdout, "corrected ctl_adr to %x\n", u.q.ctrl_adr);
     result = 1;
   }
   if (((u2.q.zero) != 0)) {
-if (DEBUG_HEUR) fprintf(stderr, "zero = %x\n", u2.q.zero);
+if (DEBUG_HEUR) fprintf(stdout, "zero = %x\n", u2.q.zero);
     u.q.zero = 0;
     result = 1;
   }
+  if (crcres && result == 0) {
+    crcres = flip_error_corr(u.b, crcres);
+    if (crcres != 0) {
+      crc_errors++;
+    }
+  }
   return result;
 }
 
+static unsigned char sectbuff[CD_FRAMESIZE_RAW + SUBSIZE];
+
 int main(int argc, char **argv)
 {
   FILE *fp;
-  unsigned char subs[SUBSIZE];
+  int i;
+  unsigned char *subs;
   unsigned last_asec = 0xffffffffU;
   unsigned last_rsec = 0xffffffffU;
   unsigned count_sectors = 0;
   unsigned sectors_trackrel = 0;
-  unsigned char tr=255, in=255;
+  unsigned char tr=1, in=0;
   int with_data = 1;
   unsigned count_rotten_entries = 0;
 
@@ -420,14 +482,13 @@ int main(int argc, char **argv)
     int crcres = 0;
 
     if (with_data)
-#if defined (SUN)
-    { unsigned char dummy[CD_FRAMESIZE_RAW];
-      fread(dummy, CD_FRAMESIZE_RAW, 1, fp);
+    {
+      fread(sectbuff, CD_FRAMESIZE_RAW + SUBSIZE, 1, fp);
+      subs = sectbuff + CD_FRAMESIZE_RAW;
+    } else {
+      fread(sectbuff, SUBSIZE, 1, fp);
+      subs = sectbuff;
     }
-#else
-      fseek (fp, CD_FRAMESIZE_RAW, SEEK_CUR); /* skip cdda data */
-#endif
-    fread (subs, SUBSIZE, 1, fp);
 
     crcres = getQ(subs);
 
@@ -435,16 +496,27 @@ int main(int argc, char **argv)
 
     /* ignore anything beside position info */
     if (crcres == 0 && (u.q.ctrl_adr & 0x0f) != 1) {
-      last_asec++; last_rsec++;
+      last_asec++;
+      if (in == 0) {
+        last_rsec--;
+      } else {
+        last_rsec++;
+      }
       continue;
     }
 
-    if (crcres && correct_heuristically(u, tr, in, last_asec, last_rsec)) {
-      crcres = calcCRC(u.b, SUBSIZE/8);
-      if (0 == crcres) count_rotten_entries--;
-    }
 #if 1
+    for( i = 1; i >= 0 ; i--) {
+      correct_heuristically(u, tr, in, last_asec, last_rsec, crcres);
+      crcres = calcCRC(u.b, SUBSIZE/8);
+      if (0 == crcres) {
+        count_rotten_entries--;
+        break;
+      }
+    }
+#if 0
     if (0 != crcres) continue;  /* skip invalid frames */
+#endif
 #endif
 
     bcd2bin(&u.q.track);
@@ -463,7 +535,7 @@ int main(int argc, char **argv)
       FILE *outchannel;
 
       if (0 != crcres)
-        outchannel = stderr;
+        outchannel = stdout;
       else {
         outchannel = stdout;
         if (tr != u.q.track) {
@@ -472,8 +544,8 @@ int main(int argc, char **argv)
       }
 
       fprintf(outchannel,
-      "%s OK-"
-      "sec: %6u   MSF %2d:%2d.%02d  rel_sec: %6u   MSF %2d:%2d.%02d    t %2d, i %2d\n",
+      "%s OK "
+      "sec: %6u   MSF %2d:%02d.%02d  rel_sec: %6u   MSF %2d:%02d.%02d    t %2d, i %2d\n",
       crcres ? "NOT" : "   ",
       u.q.aframe + (75*u.q.asec) + (75*60*u.q.amin) - 150,
       u.q.amin, u.q.asec, u.q.aframe,
@@ -481,14 +553,14 @@ int main(int argc, char **argv)
       u.q.min, u.q.sec, u.q.frame,
       u.q.track, u.q.index);
 
-      if (0 == crcres) {
+      if (1 || 0 == crcres) {
         tr = u.q.track;
         in = u.q.index;
       }
     }
   }
-  if (count_sectors && count_rotten_entries)
+  if (count_sectors && crc_errors)
     fprintf(stdout, "%u sectors with errors of %u sectors total.\n",
-            count_rotten_entries, count_sectors);
+            crc_errors, count_sectors);
   return 0;
 }
