@@ -22,7 +22,7 @@
 #include <scsireg.h>
 #include <scsitransp.h>
 
-#include "readall.h"
+#include "read2448.h"
 #include "interface.h"
 #include "byteorder.h"
 #include "global.h"
